/*******************************************************************
 Copyright (C) 2004 Thomas Kunz, CRC Canada, BCAST for IPv4.
 DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 See the GNU Library General Public License (file COPYING in the 
 distribution) for conditions of use and redistribution.
*********************************************************************/

/* 
 * Demo client application to send multicast data stream.
 * The application takes four parameters: multicast address,
 * inter-packet time in milliseconds, total number of data 
 * packets to be transmitted, and packet size (in bytes). 
 * Each data packet will start with a specific BCAST header 
 * that includes the sender's address and a sequence number.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <linux/sockios.h>
#include <sys/time.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h> /* close */

#include "bcast.h"

#define SERVER_PORT 2500  // could really be any port
#define MAX_TTL 10	  // to limit uncontrolled flooding
#define MAX_NAME_LEN 50   // host names should not be longer

int main(int argc, char *argv[]) {

  int sd, rc, i;
  int sock;
  int total_packets = 0;
  int packet_interarrival_time = 0;
  int packet_size = 0;
  unsigned char ttl = MAX_TTL;
  char host_name[MAX_NAME_LEN];
  char *data_packet;
  struct ifreq ifreq;
  struct bcastHeader bcast_hdr;
  struct sockaddr_in cliAddr, servAddr;
  struct sockaddr_in *saptr = NULL;
  struct hostent *h;
  struct timeval tv;

  if(argc != 5) {
    printf("Usage: %s <IPv4 mcast address> <time(ms)> <#packets> <size(bytes)>\n", argv[0]);
    exit(1);
  }

  /* interpret the command-line parameters */
  packet_interarrival_time = atoi(argv[2]);
  total_packets = atoi(argv[3]);
  packet_size = atoi(argv[4]);
  /* add the BCAST header */
  packet_size += sizeof(struct bcastHeader);

  /* get address of sender */
  // need a socket for ioctl()
  if( (sock = socket(AF_INET, SOCK_STREAM, 0)) < 0){
    perror("socket");
    exit(1);
  }

  strncpy(ifreq.ifr_name, BCAST_INTERFACE, IF_NAMESIZE);

  // get the address for this interface
  if (ioctl(sock, SIOCGIFADDR, &ifreq) != 0){
    perror("ioctl");
    exit(1);
  }
  // print out the address
  saptr = (struct sockaddr_in *)&ifreq.ifr_addr;
#ifdef DEBUG
  printf( "%s::%3d.%3d.%3d.%3d\n",
    		ifreq.ifr_name,
                NIPQUAD(saptr->sin_addr.s_addr) );
#endif
  memcpy((char *)&bcast_hdr.sender_addr, (char *)&saptr->sin_addr.s_addr, 4);
  bcast_hdr.prev_hop = bcast_hdr.sender_addr; // trivially true for sender

  /* prepare to create sockets */
  h = gethostbyname(argv[1]);
  if(h == NULL) {
    printf("%s : unknown host '%s'\n",argv[0],argv[1]);
    exit(1);
  }

  servAddr.sin_family = h->h_addrtype;
  memcpy((char *) &servAddr.sin_addr.s_addr, h->h_addr_list[0],h->h_length);
  servAddr.sin_port = htons(SERVER_PORT);
  
  /* check if dest address is multicast */
  if(!IN_MULTICAST(ntohl(servAddr.sin_addr.s_addr))) {
    printf("%s : given address '%s' is not multicast \n",argv[0],
	   inet_ntoa(servAddr.sin_addr));
    exit(1);
  }

  /* create socket */
  sd = socket(AF_INET,SOCK_DGRAM,0);
  if (sd < 0) {
    printf("%s : cannot open socket\n",argv[0]);
    exit(1);
  }
  
  /* bind any port number */
  cliAddr.sin_family = AF_INET;
  cliAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  cliAddr.sin_port = htons(0);
  if(bind(sd,(struct sockaddr *) &cliAddr,sizeof(cliAddr))<0) {
    perror("bind");
    exit(1);
  }

  if(setsockopt(sd,IPPROTO_IP,IP_MULTICAST_TTL,&ttl,sizeof(ttl)) < 0) {
    printf("%s : cannot set ttl = %d \n",argv[0],ttl);
    exit(1);
  }

#ifdef DEBUG
  printf("%s : sending data on multicast group '%s' (%s)\n",argv[0],
	 h->h_name,inet_ntoa(*(struct in_addr *) h->h_addr_list[0]));
#endif
  
  /* prepare data packet */
  data_packet = (char *)malloc(packet_size);
  if (data_packet == NULL) {
    printf("%s : cannot allocate %d bytes for message\n", argv[0], packet_size);
    exit(1);
  }

  for (i = 0; i < packet_size; i++) {
    data_packet[i] = '\0';
  }
  bcast_hdr.type = BCAST_DATA_PACKET;

  /* send data packets */
  for (i = 0; i < total_packets; i++) {
    bcast_hdr.sequence_no = i;
    bcopy(&bcast_hdr, data_packet, sizeof(struct bcastHeader));
    rc = sendto(sd, data_packet, packet_size, 0,
		(struct sockaddr *) &servAddr, sizeof(servAddr));
    
    if (rc < 0) {
      printf("%s : cannot send data %d\n",argv[0],i-1);
      close(sd);
      exit(1);
    }

    /* sleep for a while */
    tv.tv_sec = 0;
    tv.tv_usec = packet_interarrival_time * 1000;
    (void) select(0, NULL, NULL, NULL, &tv);

  }/* end for */

  /* close socket and exit */
  close(sd);
  exit(0);
}
