/*******************************************************************
 Copyright (C) 2004 Thomas Kunz, CRC Canada, BCAST for IPv4.
 DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 See the GNU Library General Public License (file COPYING in
 directory application) for conditions of use and redistribution.
*********************************************************************/

// Core implementation of BCAST

#include "bcast_module.h"
#include "bcast_tgt.hh"

#include <math.h>
#include <pcap.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <linux/sockios.h>
#include <linux/netfilter.h>
#include <net/if.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <list>

#define DEBUG_HELLO
#define DEBUG_NACK
#define DEBUG_MSG
#define DEBUG_REL
#define DEBUG_ALL

#define CURRENT_TIME(tv)    (void) gettimeofday(&tv, NULL)

// ******************************************************
// helper function to periodically transmit HELLO message
// ******************************************************

void send_hello_message(XrlBcastTarget *obj)
{ struct bcastHeader bcast_hdr;
  int no_entries = obj->nbs.neighbors.size();      // no of entries in message
  int msg_size = no_entries * sizeof(in_addr_t);   // multiplied by entry size
  msg_size = msg_size + sizeof(int);  // first entry is number of entries
  msg_size = msg_size + sizeof(struct bcastHeader);// protocol header

  // check that we do not exceed max size
  if (msg_size > BCAST_MAX_MSG) {
    fprintf(stderr, "BCAST: message size %d exceeds MAX of %d\n", 
      msg_size, BCAST_MAX_MSG);
    exit(1);
  }

  // allocate memory to hold message
  char *outdata = (char *)malloc(msg_size);
  if (outdata == NULL) {
    fprintf(stderr, "BCAST: cannot allocate memory to build HELLO message\n");
    exit(1);
  }

  // first entry: BCAST Header with TYPE = BCAST_HELLO_MSG and SENDER_ADDR
  // our own address, set the other fields for the time being
  bcast_hdr.type = BCAST_HELLO_MSG;
  bcast_hdr.sender_addr = obj->sender_addr;
  bcast_hdr.prev_hop = 0;
  bcast_hdr.sequence_no = 0;
  memcpy(outdata, (char *)&bcast_hdr, sizeof(struct bcastHeader));

  // second entry: number of neighbor entries
  char *next = outdata + sizeof(struct bcastHeader);
  memcpy(next, (char *) &no_entries, sizeof(int));
  next += sizeof(int);

#ifdef DEBUG_HELLO
  fprintf(stderr, "BCAST: Hello from %3d.%3d.%3d.%3d with %d entries:", 
    NIPQUAD(obj->sender_addr), no_entries);
#endif

  // next entries: known neighbors
  list<Neighbor_entry>::iterator iter;
  for (iter=obj->nbs.neighbors.begin();iter!=obj->nbs.neighbors.end();iter++) {
    memcpy(next, (char *)&(*iter).nb_addr, sizeof(in_addr_t));
    next += sizeof(in_addr_t);   
#ifdef DEBUG_HELLO
    fprintf(stderr, " %3d.%3d.%3d.%3d; ", NIPQUAD((*iter).nb_addr));
#endif
  }

#ifdef DEBUG_HELLO
  // print out time of next HELLO message
  struct timeval tv;
  (void) gettimeofday(&tv, NULL);
  fprintf(stderr, "\nBCAST: send HELLO at %d secs, %d usec\n", 
    (int)tv.tv_sec, (int)tv.tv_usec);
#endif

  int rc = sendto(obj->bcast_out_sock, outdata, msg_size, 0,
		(struct sockaddr *) &(obj->servAddr), sizeof(obj->servAddr));
 
  if (rc < 0) {
    fprintf(stderr, "BCAST: cannot send HELLO\n");
  }

  // release memory occupied by message
  free(outdata);

  // schedule next HELLO message
  double delay = (double)(random())/RAND_MAX * 
                   (MaxHelloScaling - MinHelloScaling) + MinHelloScaling;
  delay = delay * obj->hello_interval * 1000; // HELLO_INTERVAL is in seconds
  obj->hello_t = obj->bcast_events->new_oneoff_after_ms((int) delay,
                   callback(send_hello_message, obj));
}

// ****************************************************
// helper function to transmit NACK message if required
// ****************************************************

void send_nack_message(XrlBcastTarget *obj, int count,
                       in_addr_t src, u_int32_t seq_no)
{ struct bcastHeader bcast_hdr;

#ifdef DEBUG_REL
  struct timeval tv;
  CURRENT_TIME(tv);
  fprintf(stderr, "BCAST: NACK from %3d.%3d.%3d.%3d for (%3d.%3d.%3d.%3d:%d)", 
    NIPQUAD(obj->sender_addr), NIPQUAD(src), seq_no);
  fprintf(stderr, " at (%ld, %ld)\n", tv.tv_sec, tv.tv_usec);
#endif

  if (!obj->reliable_version) {
    fprintf(stderr, "BCAST: trying to send NACK, but not in reliable mode\n");
  }

  // entry: BCAST Header with TYPE = BCAST_NACK, SENDER_ADDR = SRC,
  //        SEQUENCE_NO = SEQ_NO, and PREV_HOP = own addr
  bcast_hdr.type = BCAST_NACK;
  bcast_hdr.sender_addr = src;
  bcast_hdr.sequence_no = seq_no;
  bcast_hdr.prev_hop = obj->sender_addr;

  // first check whether we received the packet, if so, silently quit
  if (obj->uids.id_lookup(bcast_hdr)) {
#ifdef DEBUG_REL
    fprintf(stderr, "BCAST: packet was received, do not send NACK\n");
#endif
    return;
  }

  // second check whether we already initiated the maximum number of NACKs
  if (count == MAX_NACK) {
#ifdef DEBUG_REL
    fprintf(stderr, "BCAST: NACK reached maximum (%d), so stop\n", MAX_NACK);
#endif
    return;
  }

  // otherwise, send NACK control message
  int rc = sendto(obj->bcast_out_sock, &bcast_hdr, sizeof(bcastHeader), 0,
		(struct sockaddr *) &(obj->servAddr), sizeof(obj->servAddr));
 
  if (rc < 0) {
    fprintf(stderr, "BCAST: cannot send NACK\n");
  }
  // reschedule the function to retransmit NACKs if necessary
  obj->nack_t = obj->bcast_events->new_oneoff_after_ms((int) NACK_TIMEOUT,
                   callback(send_nack_message, obj, count+1, src, seq_no));
}

// ******************************************************
// helper function to receive and process CONTROL message
// ******************************************************

void recv_control_message(XrlBcastTarget *obj)
{ char msg[BCAST_MAX_MSG];
  char *next;
  struct bcastHeader bcast_hdr;
  struct timeval tv;

#ifdef DEBUG_ALL
  fprintf(stderr, "BCAST: RECV control message\n");
#endif

  int rc = recvfrom(obj->bcast_in_sock, msg, BCAST_MAX_MSG, 0, NULL, 0);
  if (rc < 0) {
    fprintf(stderr, "BCAST: cannot receive data on BCAST_IN_SOCK\n");
  }
  int msg_size, expiry;
  in_addr_t temp_addr;
  list<in_addr_t> nbs;

  // extract BCAST Header
  memcpy((char*)&bcast_hdr, msg, sizeof(struct bcastHeader));
  if (bcast_hdr.type == BCAST_DATA_PACKET) {
#ifdef DEBUG_ALL
    fprintf(stderr, "BCAST: silently discarding DATA Packet\n");
#endif
    return;
  }

  // memorize time of this packet reception
  CURRENT_TIME(tv);
  obj->add_packet_arrival(tv);

  switch (bcast_hdr.type) {

    case BCAST_HELLO_MSG:
      // extract information
      next = msg + sizeof(struct bcastHeader);
      memcpy((char *)&msg_size, next, sizeof(int));
      next += sizeof(int);
#ifdef DEBUG_HELLO
      fprintf(stderr, "BCAST: rec. HELLO with %d entries from %3d.%3d.%3d.%3d:",
        msg_size, NIPQUAD(bcast_hdr.sender_addr));
#endif
      // ignore own HELLO message
      if (bcast_hdr.sender_addr == obj->sender_addr) {
#ifdef DEBUG_HELLO
        fprintf(stderr, " (droppped)\n");
#endif
        return;
      }
      for (int i = 0; i < msg_size; i++) {
        memcpy((char *)&temp_addr, next, sizeof(in_addr_t));
        next += sizeof(in_addr_t);
        nbs.insert(nbs.end(), temp_addr);
#ifdef DEBUG_HELLO
        fprintf(stderr, " %3d.%3d.%3d.%3d; ", NIPQUAD(temp_addr));
#endif
      }
#ifdef DEBUG_HELLO
      fprintf(stderr, "\n");
#endif
      expiry = obj->hello_interval * (ALLOWED_HELLO_LOSS + 1);
      obj->nbs.nb_insert(bcast_hdr.sender_addr, nbs, expiry);
      return;

    case BCAST_NACK:
#ifdef DEBUG_REL
      fprintf(stderr, "BCAST: rec. NACK from %3d.%3d.%3d.%3d", 
        NIPQUAD(bcast_hdr.prev_hop));
      fprintf(stderr, " for (%3d.%3d.%3d.%3d:%d)\n", 
        NIPQUAD(bcast_hdr.sender_addr), bcast_hdr.sequence_no);
#endif
      if (!obj->reliable_version) {
        fprintf(stderr, "BCAST: received NACK in non-reliable mode\n");
        return;
      }
      // schedule retransmission of the NACK'd packet, if we
      // have a copy in the cache
      obj->cache.init_send(bcast_hdr.sender_addr, bcast_hdr.sequence_no);
      return;

    default:
      fprintf(stderr, "BCAST: Type in BCAST_HDR (%d) is unknown\n", 
        bcast_hdr.type);
      exit(1);
  }
}

// ****************************************************************
// callback function for PCAP_DISPATCH, called whenever we captured
// a multicast packet
// ****************************************************************

void pcap_callback(u_char *args, const struct pcap_pkthdr* pkthdr,
                   const u_char* packet)
{
  const struct ether_header *eptr;  /* net/ethernet.h */
  u_int16_t type;
  in_addr_t dst_addr;
  u_int16_t dst_port;
  XrlBcastTarget *obj;
  char *payload;

  /* the user-passed parameter is a reference to an XrlBcastTarget */
  obj = (XrlBcastTarget *)args;

  /* lets start with the ether header... */
  eptr = (const struct ether_header *)packet;
  type = ntohs(eptr->ether_type);

  if (type == ETHERTYPE_IP) {
    const struct my_ip *ip;
    struct bcastHeader bcast_hdr, bcast_hdr2;
    u_int32_t length = pkthdr->len;
    int len, hlen, version;

    /* jump pass the ethernet header */
    ip = (const struct my_ip*)(packet + sizeof(struct ether_header));
    length -= sizeof(struct ether_header); 

    /* check to see we have a packet of valid length */
    if (length < sizeof(struct my_ip)) {
      fprintf(stderr, "BCAST: Truncated IP packet captured: %d",length);
      exit(1);
    }

    len = ntohs(ip->ip_len);
    hlen    = IP_HL(ip); /* header length */
    version = IP_V(ip);  /* ip version */

    /* check version */
    if (version != 4) {
      fprintf(stderr, "BCAST: Unknown IP version %d\n", version);
      exit(1);
    }

    /* check header length */
    if (hlen < 5 ) {
      fprintf(stderr, "BCAST: Bad IP header length: %d\n", hlen);
      exit(1);
    }

    // ignore short packets (may belong to underlying multicast protocol)
    if (length <= 30 + sizeof(bcastHeader)) {
#ifdef DEBUG_MSG
      fprintf(stderr, "BCAST: short packet size (%d <= %d)\n", 
        length, 30 + sizeof(bcastHeader));
      fprintf(stderr, "   Destination IP address: %3d.%3d.%3d.%3d\n",
        NIPQUAD2(ip->ip_dst));
#endif
    } else {
      // memorize destination port and multicast group
      bcopy(&(ip->ip_dst),&dst_addr, sizeof(dst_addr));
      bcopy(packet + sizeof(struct ether_header) + hlen*4 + 2, 
            &dst_port, sizeof(dst_port));

      // memorize time of this packet reception
      struct timeval tv;
      CURRENT_TIME(tv);
      obj->add_packet_arrival(tv);

      // extract BCAST header: after Ethernet, IP, and UDP header
      bcopy(packet + sizeof(struct ether_header) + hlen*4 + 8, 
           (char *)&bcast_hdr, sizeof(bcastHeader));
#ifdef DEBUG_MSG
      fprintf(stderr, "Message (%3d.%3d.%3d.%3d: %d), length %d, type %s\n", 
        NIPQUAD(bcast_hdr.sender_addr), bcast_hdr.sequence_no, length,
        ((bcast_hdr.type == BCAST_DATA_PACKET) ? "DATA" :
          (bcast_hdr.type == BCAST_HELLO_MSG) ? "HELLO" : "NACK"));
#endif

      // if this is NOT a DATA PACKET, do not process any further (well, there
      // really should not have been any non-data packets, depending on the
      // FILTER expression in BCAST_TGT.HH
      if (bcast_hdr.type != BCAST_DATA_PACKET) {
#ifdef DEBUG_MSG
        fprintf(stderr, "Drop non-data packet\n");
#endif
        return;
      }

      // learn something from that packet reception: get all neighbors
      // of node that transmitted this packet
      list<in_addr_t> prev_hop_nbs = obj->nbs.nb_lookup(bcast_hdr.prev_hop);
      prev_hop_nbs.insert(prev_hop_nbs.begin(), bcast_hdr.prev_hop);

      // check whether we already received that packet
      if (!obj->uids.id_lookup(bcast_hdr)) {

        // save unique ID
        obj->uids.id_insert(bcast_hdr);

        // if this packet was sent from this node, ignore...
        if (bcast_hdr.sender_addr == obj->sender_addr) {
#ifdef DEBUG_MSG
          fprintf(stderr, "Message originates locally, ignore\n");
#endif
          if (obj->reliable_version) {
            // but first save copy of packet in cache, for potential retransmission
            payload = (char *)const_cast<u_char *>(packet) + 
                      sizeof(struct ether_header) + // skip Ethernet header
                      hlen*4 +                      // skip variable size IP hdr
                      8 +                           // skip UDP header
                      sizeof(struct bcastHeader);   // skip BCAST header
            length -= hlen*4 + 8 + sizeof(struct bcastHeader);
            obj->cache.add_to_cache(payload, length, dst_addr, dst_port, bcast_hdr);
          }
          return;
        }

        // in the reliable version of the protocol, check whether we 
        // received the previous packet and whether receiving the current
        // packet allows us to cancel a pending NACK'd retransmission
        //
        // however, do not bother if there is a currently a set NACK timer
        if (obj->reliable_version && !obj->nack_t.scheduled()) {
          bcopy((char *)&bcast_hdr, (char *)&bcast_hdr2, sizeof(bcastHeader));
          if (bcast_hdr2.sequence_no > 0) {
            bcast_hdr2.sequence_no--; // check previous packet
            if (!obj->uids.id_lookup(bcast_hdr2) && !obj->network_congested()) {
#ifdef DEBUG_REL
              fprintf(stderr, "Missing (%3d.%3d.%3d.%3d: %d), issue NACK\n", 
                NIPQUAD(bcast_hdr2.sender_addr), bcast_hdr2.sequence_no);
#endif
              double delay = NACK_INITIAL_DELAY * (double)(random())/RAND_MAX;
              obj->nack_t = obj->bcast_events->new_oneoff_after_ms((int)delay,
                   callback(send_nack_message, obj, 0, bcast_hdr2.sender_addr,
                            bcast_hdr2.sequence_no));
            }
          }

          // also check whether this packet reception allows us to cancel
          // a pending NACK'd retransmission
          obj->cache.cancel_send(bcast_hdr.sender_addr, bcast_hdr.sequence_no);
        }

        // check whether this node has neighbors (obj->nbs.one_hop_nbs()) that 
        // are not already covered by the known neighbors of the transmitting
        // node (prev_hop_nbs)
        list<in_addr_t> my_nbs = obj->nbs.one_hop_nbs();
        list<in_addr_t>::iterator prev_nbs_iter;
        list<in_addr_t>::iterator tmp_nb_iter, nb_iter;
#ifdef DEBUG_MSG
        fprintf(stderr, "Local neighbors: ");
        for (nb_iter = my_nbs.begin(); nb_iter != my_nbs.end(); nb_iter++) {
          fprintf(stderr, "%3d.%3d.%3d.%3d; ", NIPQUAD(*nb_iter));
        }
        fprintf(stderr, "\nRemote neighbors: ");
        for (nb_iter = prev_hop_nbs.begin(); 
             nb_iter != prev_hop_nbs.end(); nb_iter++) {
          fprintf(stderr, "%3d.%3d.%3d.%3d; ", NIPQUAD(*nb_iter));
        }
        fprintf(stderr, "\n");
#endif
        for (prev_nbs_iter = prev_hop_nbs.begin(); 
             prev_nbs_iter != prev_hop_nbs.end(); prev_nbs_iter++) {
          nb_iter = my_nbs.begin();
          while (nb_iter != my_nbs.end()) {
            tmp_nb_iter = nb_iter;
            tmp_nb_iter++; // advance to next element in list
            if (*prev_nbs_iter == *nb_iter) {
#ifdef DEBUG_MSG
              fprintf(stderr, "BCAST: del %3d.%3d.%3d.%3d (%3d.%3d.%3d.%3d, %d)\n",
                NIPQUAD(*nb_iter),NIPQUAD(bcast_hdr.sender_addr),
                bcast_hdr.sequence_no);
#endif
              my_nbs.erase(nb_iter); // invalidates NB_ITER, reset from TMP_NB_ITER
            }
            nb_iter = tmp_nb_iter;
          }
          if (my_nbs.empty()) {
#ifdef DEBUG_MSG
            fprintf(stderr, "BCAST: no additional neighbors left, \"drop\" packet\n");
#endif
            obj->buf.check_against_buffer(bcast_hdr, prev_hop_nbs);
            return;
          }
        }

        // determine random delay in ms for packet retransmission
        double delay = (double)(random())/RAND_MAX;
        int node_degree = obj->nbs.node_degree();
        if (node_degree != 0) {
          delay = delay * obj->nbs.max_degree()/obj->nbs.node_degree();
        }
        delay = delay * BCAST_DELAY;
#ifdef DEBUG_MSG
        fprintf(stderr, "Add to queue with delay %lf\n", delay);
#endif

        // add packet to buffer with appropriate delay
        // skip all the various headers and only save the payload
        payload = (char *)const_cast<u_char *>(packet) + 
                  sizeof(struct ether_header) + // skip Ethernet header
                  hlen*4 +                      // skip variable size IP hdr
                  8 +                           // skip UDP header
                  sizeof(struct bcastHeader);   // skip BCAST header
        length -= hlen*4 + 8 + sizeof(struct bcastHeader);
        obj->buf.add_to_buffer(payload, length, dst_addr, dst_port, bcast_hdr, delay,
                               obj->nbs.one_hop_nbs());

        // also save copy of packet in cache, for potential retransmission
        if (obj->reliable_version) {
          obj->cache.add_to_cache(payload, length, dst_addr, dst_port, bcast_hdr);
        }
      } else {
        // silently drop packet
#ifdef DEBUG_MSG
        fprintf(stderr, "Silently drop packet (duplicate)\n");
#endif
      }

      // now use information we saved before to reduce not-yet-reached
      // neighbor list where appropriate
      obj->buf.check_against_buffer(bcast_hdr, prev_hop_nbs);
    } 
  }
}

// *****************************************************************
// helper function to poll for input messages and schedule reception
// *****************************************************************

bool poll_sockets(XrlBcastTarget *obj)
{ fd_set rfds;
  struct timeval tv;

  /* check for control messages on BCAST_IN_SOCK first:
     create FD_SET and initialize TV to ZERO (for polling) */

  FD_ZERO(&rfds);
  FD_SET(obj->bcast_in_sock, &rfds);
  tv.tv_sec = 0;
  tv.tv_usec = 0;

  int rc = select(obj->bcast_in_sock+1, &rfds, NULL, NULL, &tv);
  if (rc < 0) {
    fprintf(stderr, "BCAST: SELECT returned with error\n");
  }
  if (rc != 0) {
    /* schedule reception of appropriate message (control messages or data) */
    if (FD_ISSET(obj->bcast_in_sock, &rfds)) {
      recv_control_message(obj);
    }
  }

  // also check for captured multicast packets...
  // for non-blocking devices, have to use PCAP_DISPATCH with a callback
       
  if (pcap_dispatch(obj->pcap_descr, 1, pcap_callback, (u_char *)obj) == -1) {
    pcap_perror(obj->pcap_descr, NULL);
  }
  return true;
}

// *****************************************************
// first the methods that implement the COMMON interface
// *****************************************************

XrlCmdError
XrlBcastTarget::common_0_1_get_target_name(string& name)
{
    name = "bcast";
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlBcastTarget::common_0_1_get_version(string& version)
{
    version = "1.0";
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlBcastTarget::common_0_1_get_status(uint32_t& status, 
				     string& reason)
{
    status = status_code;
    reason = "None";
    return XrlCmdError::OKAY();
}

XrlCmdError
XrlBcastTarget::common_0_1_shutdown()
{
    // closed the various file descriptors
    close(bcast_in_sock);
    close(bcast_out_sock);
    pcap_close(pcap_descr);;
    // status_code = PROC_SHUTDOWN;
    exit(0);
    // return XrlCmdError::OKAY();
}

// *****************************************************
// second the methods that implement the BCAST interface
// *****************************************************

XrlCmdError XrlBcastTarget::bcast_1_0_set_hello_interval(
	// Input values, 
	const int32_t&	new_int)
{
    hello_interval = new_int;
    return XrlCmdError::OKAY();
}

XrlCmdError XrlBcastTarget::bcast_1_0_get_hello_interval(
	// Output values, 
	int32_t&	old_int)
{
    old_int = hello_interval;
    return XrlCmdError::OKAY();
}

XrlCmdError XrlBcastTarget::bcast_1_0_set_buffer_size(
	// Input values, 
	const int32_t&	new_buf)
{
    buffer_size = new_buf;
    return XrlCmdError::OKAY();
}

XrlCmdError XrlBcastTarget::bcast_1_0_get_buffer_size(
	// Output values, 
	int32_t&	old_buf)
{
    old_buf = buffer_size;
    return XrlCmdError::OKAY();
}

XrlCmdError XrlBcastTarget::bcast_1_0_set_reliability(
	// Input values, 
	const bool&	reliability)
{
    reliable_version = reliability;
    return XrlCmdError::OKAY();
}

XrlCmdError XrlBcastTarget::bcast_1_0_get_reliability(
	// Output values, 
	bool&	reliability)
{
    reliability = reliable_version;
    return XrlCmdError::OKAY();
}

XrlCmdError XrlBcastTarget::bcast_1_0_set_interface(
	// Input values, 
	const string&	intf)
{
    interface = intf;
    return XrlCmdError::OKAY();
}

XrlCmdError XrlBcastTarget::bcast_1_0_get_interface(
	// Output values, 
	string&	intf)
{
    intf = interface;
    return XrlCmdError::OKAY();
}

XrlCmdError XrlBcastTarget::bcast_1_0_activate_bcast()
{
    status_code = PROC_READY;

    struct ip_mreq mreq;
    struct hostent *h;
    struct ifreq ifreq;

    char errbuf[PCAP_ERRBUF_SIZE];
    struct bpf_program fp;      /* PCAP: hold compiled program     */
    bpf_u_int32 maskp;          /* PCAP: subnet mask               */
    bpf_u_int32 netp;           /* PCAP: ip                        */

    int sock;

    // we have to do a whole bunch of initialization:
    // 1) create a socket to receive messages from (bcast_in_sock)
    // 2) create a socket to send messages out (with TTL set to 1)
    // 3) determine a UNIQUE ID for this instance: the IP address of the
    //    protocol interface (which is set via XRL on startup and initialized
    //    to BCAST_INTERFACE
    // 4) set up the pcap stuff to intercept packets
    // 5) schedule the first transmission of hello messages
    // 6) schedule the periodic polling of INPUT file descriptors
    // 7) schedule the first test for expired neighbors
    // 8) schedule the first test for expired unique PACKET IDs

    // 0) some sanity check on config info for control multicast group

    /* get mcast address to listen to */
    h = gethostbyname(BCAST_GROUP);
    if (h == NULL) {
      return XrlCmdError::COMMAND_FAILED("BCAST: unknown multicast group");
    }
  
    memcpy(&mcastAddr, h->h_addr_list[0],h->h_length);
 
    /* check given address is multicast */
    if (!IN_MULTICAST(ntohl(mcastAddr.s_addr))) {
      return XrlCmdError::COMMAND_FAILED("BCAST: address is not multicast");
    }

    // initialize the one known field in destAddr
    destAddr.sin_family = AF_INET;

    /* 1) create receiving control socket */
    bcast_in_sock = socket(AF_INET,SOCK_DGRAM,0);
    if (bcast_in_sock < 0) {
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot create BCAST_IN_SOCK");
    }

    /* bind port */
    servAddr.sin_family = AF_INET;
    servAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servAddr.sin_port = htons(BCAST_PORT);  
    if (bind(bcast_in_sock,(struct sockaddr *)&servAddr,sizeof(servAddr)) < 0) {
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot bind BCAST_IN_SOCK");
    }
    /* replace INADDR_ANY in servAddr with the multicast address */
    servAddr.sin_addr.s_addr = mcastAddr.s_addr;

    /* join multicast group */
    mreq.imr_multiaddr.s_addr = mcastAddr.s_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
  
    int rc = setsockopt(bcast_in_sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		    (void *) &mreq, sizeof(mreq));
    if (rc < 0) {
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot join multicast group");
    }
    printf("BCAST: listening to mgroup %s:%d\n", BCAST_GROUP, BCAST_PORT);

    /* 2) create control message transmission socket */
    bcast_out_sock = socket(AF_INET,SOCK_DGRAM,0);
    if (bcast_out_sock < 0) {
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot open BCAST_OUT_SOCK");
    }
  
    /* bind any port number */
    cliAddr.sin_family = AF_INET;
    cliAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    cliAddr.sin_port = htons(0);
    if(bind(bcast_out_sock,(struct sockaddr *) &cliAddr, sizeof(cliAddr)) < 0) {
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot bind BCAST_OUT_SOCK");
    }

    unsigned char ttl = BCAST_TTL;
    if (setsockopt(bcast_out_sock, IPPROTO_IP, IP_MULTICAST_TTL,
	&ttl, sizeof(ttl)) < 0) {
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot set TTL");
    }

    // 3) determine IP address of protocol interface
    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0){
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot create SOCK");
    }

    strncpy(ifreq.ifr_name, interface.c_str(), IF_NAMESIZE);
    // get the address for this interface
    if (ioctl(sock, SIOCGIFADDR, &ifreq) != 0){
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot perform IOCTL on SOCK");
    }
    struct sockaddr_in *saptr = (struct sockaddr_in *)&ifreq.ifr_addr;
    memcpy((char *)&sender_addr, 
	   (char *)&saptr->sin_addr.s_addr,
	   sizeof(in_addr_t));

    // 4) initialize PCAP filter to snoop on multicast packets
    /* ask pcap for the network address and mask of the device */
    pcap_lookupnet(const_cast<char *>(interface.c_str()), 
      &netp, &maskp, errbuf);

    /* open device for reading, set it in promiscuous
     * mode so we can monitor traffic to another machine */
    pcap_descr = pcap_open_live(const_cast<char *>(interface.c_str()), 
      BUFSIZ, 1, -1, errbuf);
    if (pcap_descr == NULL) {
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot open PCAP_DESCR");
    }

    /* Lets try and compile the program.. non-optimized */
    if (pcap_compile(pcap_descr,&fp,const_cast<char *>(FILTER),0,netp) == -1) { 
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot compile PCAP filter");
    }

    /* set the compiled program as the filter */
    if (pcap_setfilter(pcap_descr, &fp) == -1) { 
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot set PCAP filter");
    }

    /* make device non-blocking */
    if (pcap_setnonblock(pcap_descr, 1, errbuf) == -1) {
      return XrlCmdError::COMMAND_FAILED("BCAST: cannot make PCAP nonblock");
    }

    // 5) schedule first transmission of HELLO message
    double delay = (double)(random())/RAND_MAX *
                     (MaxHelloScaling - MinHelloScaling) + MinHelloScaling;
    delay = delay * hello_interval * 1000;
    hello_t = bcast_events->new_oneoff_after_ms((int) delay,
		callback(send_hello_message, this));

    // 6) poll reception sockets every POLL milliseconds
    poll_t = bcast_events->new_periodic(BCAST_POLL_PERIOD,
                        callback(poll_sockets, this));

    // 7) schedule first test for expired neighbour information
    int nb_delay = hello_interval * ALLOWED_HELLO_LOSS * 1000; 
    neighbor_t = bcast_events->new_oneoff_after_ms(nb_delay,
                     callback(handle_neighbors, this, &nbs));

    // 8) schedule first cleanup of expired unique packet IDs
    cleanup_t = bcast_events->new_oneoff_after_ms(CLEANUP_PERIOD,
                   callback(handle_cleanup, this, &uids));

    // once we are here, all initialization is done :-)
    return XrlCmdError::OKAY();
}

// *******************************************************
// here are the private BCAST methods and helper functions
// *******************************************************

void XrlBcastTarget::add_packet_arrival(struct timeval t)
{ // add time to the end of the list, but make sure the total
  // size does not exceed PACKET_COUNT. If so, delete front of list
  if (packet_arrivals.size() == PACKET_COUNT) 
    packet_arrivals.erase(packet_arrivals.begin());
  packet_arrivals.insert(packet_arrivals.end(), t);
}

bool XrlBcastTarget::network_congested()
{ // if the time difference between the last (i.e., most recent) and first
  // timestamp in the last does not exceed MIN_DIFF, there is too much
  // traffic in the network, so we return TRUE to prevent further congesting
  // the network with NACKs and ensuing packet retransmissions.
  struct timeval result;
  if (packet_arrivals.empty() || packet_arrivals.size() < PACKET_COUNT) {
    return FALSE; // trivially wrong
  } else {
    timersub(&packet_arrivals.back(), &packet_arrivals.front(), &result);
#ifdef DEBUG_NACK
    fprintf(stderr, "BCAST: check for congestion, time diff: %d %d", 
      (int)result.tv_sec, (int)result.tv_usec);
    fprintf(stderr, " %s\n", 
      (result.tv_sec == 0 && result.tv_usec <= MIN_DIFF)?"rejected":"approved");
#endif
    if (result.tv_sec != 0 || result.tv_usec > MIN_DIFF) {
      return FALSE;
    } else {
      return TRUE;
    }
  }
}


