/*******************************************************************
 Copyright (C) 2004 Thomas Kunz, CRC Canada, BCAST for IPv4.
 DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 See the GNU Library General Public License (file COPYING in
 directory application) for conditions of use and redistribution.
*********************************************************************/

// Core implementation of BCAST

#include <pcap.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <list>

#include "bcast.hh"
#include "bcast_const.hh"
#include "libxorp/status_codes.h"
#include "libxorp/timer.hh"
#include "libxipc/xrl_router.hh"
#include "xrl/targets/bcast_base.hh"

// For each instance of BCAST, we have objects that
// - track unique packet IDs for a while: UIDS
// - track 1-hop and 2-hop neighbor information: NBS
// - implement a pending packet buffer (packets yet to be transmitted): BUF
// - implement a packet chache of recently transmitted packets: CACHE

#include "unique_id.hh"
#include "neighbors.hh"
#include "buffer.hh"
#include "cache.hh"

// formatting macro for IPv4 addresses

#define NIPQUAD(addr)   ((unsigned char *)&addr)[0], \
				((unsigned char *)&addr)[1], \
				((unsigned char *)&addr)[2], \
				((unsigned char *)&addr)[3]

#define NIPQUAD2(addr)  ((const unsigned char *)&addr)[0], \
				((const unsigned char *)&addr)[1], \
				((const unsigned char *)&addr)[2], \
				((const unsigned char *)&addr)[3]

// struct to extract IPv4 header info

struct my_ip {
        u_int8_t        ip_vhl;         /* header length, version */
#define IP_V(ip)        (((ip)->ip_vhl & 0xf0) >> 4)
#define IP_HL(ip)       ((ip)->ip_vhl & 0x0f)
        u_int8_t        ip_tos;         /* type of service */
        u_int16_t       ip_len;         /* total length */
        u_int16_t       ip_id;          /* identification */
        u_int16_t       ip_off;         /* fragment offset field */
#define IP_DF 0x4000                    /* dont fragment flag */
#define IP_MF 0x2000                    /* more fragments flag */
#define IP_OFFMASK 0x1fff               /* mask for fragmenting bits */
        u_int8_t        ip_ttl;         /* time to live */
        u_int8_t        ip_p;           /* protocol */
        u_int16_t       ip_sum;         /* checksum */
        struct in_addr  ip_src,ip_dst;  /* source and dest address */
};

// filter expression for PCAP: grab all packets send to the BCAST control port as
// well as all packets that are sent to IP Multicast addresses not in 224.0.0.*

#define FILTER \
    "(ip multicast and not dst net 224.0.0 and not dst host 255.255.255.255)"

// the filter expression below also redirects all CONTROL packets to 
// PCAP, but we do not want to process them if we can avoid this...
//
// "(dst host 224.0.0.128 and dst port 1500) or 
//  (ip multicast and not dst net 224.0.0 and not dst host 255.255.255.255)"

/*************************************************************************
 * The class XrlBcastTarget: implement methods to deal with all XRL
 * invocations, in particular the ACTIVATE method
 *************************************************************************/

class XrlBcastTarget: public XrlBcastTargetBase {
public:
    XrlBcastTarget(EventLoop* e, XrlRouter* r) 
      : XrlBcastTargetBase(r), buf(this), cache(this)
    { bcast_events = e; 
      hello_interval = DEFAULT_HELLO_INTERVAL;
      buffer_size = DEFAULT_BUFFER_SIZE;
      reliable_version = false; 
      interface = BCAST_INTERFACE;
      status_code = PROC_READY;
    }

protected:
    // Methods to be implemented by derived classes supporting this interface.
    virtual XrlCmdError common_0_1_get_target_name(
	// Output values, 
	string&	name);

    virtual XrlCmdError common_0_1_get_version(
	// Output values, 
	string&	version);

    virtual XrlCmdError common_0_1_get_status(
	// Output values,
        uint32_t& status,
	string&	reason);

    virtual XrlCmdError common_0_1_shutdown();

    virtual XrlCmdError bcast_1_0_set_hello_interval(
	// Input values, 
	const int32_t&	new_int);

    virtual XrlCmdError bcast_1_0_get_hello_interval(
	// Output values, 
	int32_t&	old_int);

    virtual XrlCmdError bcast_1_0_set_buffer_size(
	// Input values, 
	const int32_t&	new_buf);

    virtual XrlCmdError bcast_1_0_get_buffer_size(
	// Output values, 
	int32_t&	old_buf);

    virtual XrlCmdError bcast_1_0_set_reliability(
	// Input values, 
	const bool&	reliability);

    virtual XrlCmdError bcast_1_0_get_reliability(
	// Output values, 
	bool&	reliability);

    virtual XrlCmdError bcast_1_0_set_interface(
	// Input values, 
	const string&	intf);

    virtual XrlCmdError bcast_1_0_get_interface(
	// Output values, 
	string&	intf);

    virtual XrlCmdError bcast_1_0_activate_bcast();

public:
    EventLoop *bcast_events;    // pointer to global event loop 
    XorpTimer hello_t;          // to schedule HELLO messages
    XorpTimer poll_t;           // to schedule POLLING of input sockets
    XorpTimer cleanup_t;	// to schedule CLEANUP of UNIQUE IDs
    XorpTimer neighbor_t;	// check whether neighbors expired
    XorpTimer next_msg_t;       // next pending message
    XorpTimer retransmit_t;     // delay retransmission of NACK'd packets
    XorpTimer nack_t;		// potentially retransmit NACKs
    int hello_interval;         // how often to send a HELLO message
    in_addr_t sender_addr;      // unique node ID

    UniqueIDs uids;             // recently seen packets
    Neighbors nbs;              // information about neighbors
    bcast_buffer buf;           // packets awaiting re-broadcast
    bcast_cache cache;          // cache of recently transmitted packets
    int buffer_size;            // how many recent packets should be buffered
    bool reliable_version;      // does BCAST operate with NACKs?
   
private:
    int status_code;            // captures status of BCAST, not really used much
    string interface;           // network interface name over which BCAST works
    list<struct timeval> packet_arrivals; // FIFO list of recent receptions

public:
    int bcast_out_sock;         // socket for sending messages
    int bcast_in_sock;          // socket for receiving messages
    pcap_t *pcap_descr;         // capturing multicast packets
    struct sockaddr_in cliAddr; // Client Address (source)
    struct sockaddr_in servAddr;// Server Address (destination)
    struct sockaddr_in destAddr;// Destination for individual multicast packets
    struct in_addr mcastAddr;   // To be used to test multicast destination

    void add_packet_arrival(struct timeval t); // save time of packet arrival
    bool network_congested();   // check whether network is heavily loaded,
                                // based on rate of most recent packet arrival
};
