/*******************************************************************
 Copyright (C) 2004 Thomas Kunz, CRC Canada, BCAST for IPv4.
 DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 See the GNU Library General Public License (file COPYING in
 directory application) for conditions of use and redistribution.
*********************************************************************/

/* BCAST buffers packets before they are retransmitted, delaying them
 * by a certain amount of time. The necessary functionality is implemented
 * in class BCAST_BUFFER.
 */

#include "bcast_module.h"
#include "bcast_tgt.hh"

#include <stdio.h>

#define CURRENT_TIME(tv)    (void) gettimeofday(&tv, NULL)
#define DEBUG

/***********************************************************************
 * MISCELLANEOUS: callback function to invoke transmission of first
 *                pending packet in buffer
 ***********************************************************************/

void
handle_buffer(bcast_buffer *buf) 
{ // transmit first message and reschedule next one
  buf->send_first_msg();
}

/***********************************************************************
 * Methods dealing with a single BUFFER entry
 ***********************************************************************/

// compare the known reached neighbors DESTS of the most recent message 
// identified by MSG_HDR against the list of yet-to-be-reached neighbors 
// for that entry, assuming it is the same BCAST message (same sender and 
// sequence number). Return TRUE if the list is empty, FALSE otherwise

bool bcast_buffer_entry::compare_neighbor_list
	(struct bcastHeader msg_hdr, list<in_addr_t> dests)
{ 
  // first check whether we are dealing with the same BCAST message
  if (msg_hdr.sender_addr != bcast_hdr.sender_addr ||
      msg_hdr.sequence_no != bcast_hdr.sequence_no)
    return false;

  // go through the list of newly reached neighbours in DESTS and
  // see whether that covers a neighor in the yet-to-be-reached list NEIGBOR.
  // If this is the case, delete that entry.
  list<in_addr_t>::iterator dests_iter;
  list<in_addr_t>::iterator tmp_nb_iter, nb_iter;
  for (dests_iter = dests.begin(); dests_iter != dests.end(); dests_iter++) {
    nb_iter = neighbors.begin();
    while (nb_iter != neighbors.end()) {
      tmp_nb_iter = nb_iter;
      tmp_nb_iter++; // save pointer to next element in list
      if (*dests_iter == *nb_iter) {
#ifdef DEBUG
        fprintf(stderr, "BCAST: del %3d.%3d.%3d.%3d (%3d.%3d.%3d.%3d, %d)\n",
          NIPQUAD(*nb_iter),NIPQUAD(bcast_hdr.sender_addr),
          bcast_hdr.sequence_no);
#endif
        neighbors.erase(nb_iter); // invalidates NB_ITER, reset from TMP_NB_ITER
      }
      nb_iter = tmp_nb_iter;
    }
  }
  if (neighbors.size() == 0)
    return true;
  else
    return false;
}

// check whether the new timeout value *TV for the packet identified by MSG_HDR
// would cause this packet to be retransmitted before an earlier packet from the
// same sender, which could cause packet reordering. If so, adjust *TV 
// appropriately (i.e., keep packets from same sender in sequence).

void bcast_buffer_entry::adjust_timer(struct bcastHeader msg_hdr, 
  struct timeval *tv)
{ struct timeval add_on;
  add_on.tv_sec = 0;
  add_on.tv_usec = 0;
  // first check whether we are dealing with an earlier packet from
  // the same sender
  if (msg_hdr.sender_addr == bcast_hdr.sender_addr &&
      msg_hdr.sequence_no > bcast_hdr.sequence_no) {
    // now check wether new expiry time would cause re-ordering
    if (timercmp(tv, &expiry, <)) {
      // add an additional random delay of between 0 to 2 ms
      add_on.tv_usec = (uint) (2000 * (double)(random())/RAND_MAX);
      timeradd(tv, &add_on, tv);
    }
  }
}

// send this message (typically the first one in the list of pending
// messages) 

void bcast_buffer_entry::send_msg(XrlBcastTarget *obj)
{ unsigned char msg[BCAST_MAX_MSG];

  // we have to change the entry on BCAST_HDR to list the current node
  // as previous hop
  bcast_hdr.prev_hop = obj->sender_addr;

#ifdef DEBUG
  fprintf(stderr, "BCAST: send pending entry for (%3d.%3d.%3d.%3d, %d)\n",
    NIPQUAD(bcast_hdr.sender_addr), bcast_hdr.sequence_no);
#endif

  // assemble message: first BCAST_HDR, then PAYLOAD
  bcopy((char *)&bcast_hdr, msg, sizeof(bcastHeader));
  bcopy(buf, msg + sizeof(struct bcastHeader), buf_len);

  // send to correct destination addr
  bcopy(&dst_addr, &(obj->destAddr.sin_addr), sizeof(dst_addr));
  obj->destAddr.sin_port = dst_port;

  int rc = sendto(obj->bcast_out_sock, msg, 
                  buf_len + sizeof(struct bcastHeader), 0,
		  (struct sockaddr *) &(obj->destAddr), sizeof(obj->destAddr));
 
  if (rc < 0) {
    fprintf(stderr, "BCAST: cannot send pending multicast packet\n");
    exit(1);
  }
}

/************************************************************************
 * Methods dealing with the complete packet BUFFER
 ************************************************************************/

// prepare an entry for the buffer of pending BCAST messages and add to buffer

void bcast_buffer::add_to_buffer(char *msg, int msg_len, in_addr_t addr, u_int16_t port,
  struct bcastHeader msg_hdr, double delay, list<in_addr_t> nb)
{ struct timeval now, add_tv, expire;

  // figure out when packet should be retransmitted, based on additional delay
  CURRENT_TIME(now);
  add_tv.tv_sec = 0;
  add_tv.tv_usec = (int)(delay * 1000); // delay is in ms
  timeradd(&now, &add_tv, &expire);

  // make sure we are not scheduling a packet for retransmission before
  // an earlier packet from the same sender in the buffer - if so, adjust time
  list<bcast_buffer_entry>::iterator buf_iter;
  for (buf_iter = buffer.begin(); buf_iter != buffer.end(); buf_iter++) {
    (*buf_iter).adjust_timer(msg_hdr, &expire);
  }

#ifdef DEBUG
  fprintf(stderr, "BCAST: add packet for transmission at (%d, %d)\n",
    (int)expire.tv_sec, (int)expire.tv_usec);
#endif

  // we are done: create an entry and add to BUFFER (based on timestamp value)
  bcast_buffer_entry *new_entry = new bcast_buffer_entry(msg, msg_len, addr, port,
                                                         msg_hdr, expire, nb);
  for (buf_iter = buffer.begin(); buf_iter != buffer.end(); buf_iter++) {
    if (!timercmp(&expire, &(*buf_iter).time_val(), <))
      break;
  }
  buffer.insert(buf_iter, *new_entry);
  if (buffer.size() == 1) {
    // need to schedule timer
    schedule_first_message();
  } else if (buf_iter == buffer.begin()) {
    // we inserted the packet at the beginning of the queue, re-adjust
    // timer to expire earlier now
    schedule_first_message();
  }
}

// iterate over all buffer entries and see whether we can reduce the
// number of yet-to-be-reached neighbors. For entries that have no further
// pending neigbors, delete them.

void bcast_buffer::check_against_buffer(struct bcastHeader msg_hdr, 
                                        list<in_addr_t> dests)
{ list<bcast_buffer_entry>::iterator buf_iter, tmp_buf_iter;
  buf_iter = buffer.begin();
#ifdef DEBUG
  fprintf(stderr, "BCAST: check (%3d.%3d.%3d.%3d, %d) against buffer\n",
    NIPQUAD(msg_hdr.sender_addr), msg_hdr.sequence_no);
#endif

  while (buf_iter != buffer.end()) {
    tmp_buf_iter = buf_iter;
    tmp_buf_iter++; // save a pointer to the next element
    if ((*buf_iter).compare_neighbor_list(msg_hdr, dests) == true) {
#ifdef DEBUG
      fprintf(stderr, "BCAST: del pending entry for (%3d.%3d.%3d.%3d, %d)\n",
        NIPQUAD(msg_hdr.sender_addr), msg_hdr.sequence_no);
#endif
      // we need to delete the message and the local data structures
      // check whether this is the current head-of-line packet,
      // in which case we need to reschedule the expiration timer later
      bool is_first = (buf_iter == buffer.begin());
      buffer.erase(buf_iter);    // invalidates pointer, reset from TMP_BUF_ITER

      // if we deleted the first element, need to deal with timer
      if (is_first) {
        if (!buffer.empty()) {
          schedule_first_message();
        } else {
           // if the buffer timer is set, unschedule it: no more packets
           if (bcast_obj->next_msg_t.scheduled())
             bcast_obj->next_msg_t.unschedule();
           // break out of loop: buffer is empty
           break;
        }
      }
    }
    buf_iter = tmp_buf_iter;
  }
}

// send the first message currently pending and schedule next pending message

void bcast_buffer::send_first_msg()
{ (*buffer.begin()).send_msg(bcast_obj);
  buffer.erase(buffer.begin());
  // schedule next packet in buffer, assuming there is one
  schedule_first_message();
}

// return the time at which the first currently pending message is to be sent

struct timeval& bcast_buffer::expiry_first_msg()
{ return (*buffer.begin()).time_val();
}

// schedule a callback when the head-of-line packet should be transmitted

void bcast_buffer::schedule_first_message()
{ if (!buffer.empty()) {
    // first, if the timer is currently set, unschedule it
    if (bcast_obj->next_msg_t.scheduled())
      bcast_obj->next_msg_t.unschedule();

    // create a time structure based on exiry time of head of the line
    TimeVal next((*buffer.begin()).time_val().tv_sec, 
                 (*buffer.begin()).time_val().tv_usec);

    bcast_obj->next_msg_t = bcast_obj->bcast_events->new_oneoff_at(next,
                 callback(handle_buffer, this));
  }
}
