/*******************************************************************
 Copyright (C) 2004 Thomas Kunz, CRC Canada, BCAST for IPv4.
 DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 See the GNU Library General Public License (file COPYING in
 directory application) for conditions of use and redistribution.
*********************************************************************/

/* For the reliable BCAST version, we need to buffer recently
 * transmitted packets. The required functionality is implemented in
 * the class BCAST_CACHE.
 */

#include "bcast_module.h"
#include "bcast_tgt.hh"

#include <stdio.h>

#define DEBUG

/************************************************************************
 * Callbacks for various timer-related functions
 ************************************************************************/

// when the pending retransmission timer goes off, the pending
// packet retransmission is actually transmitted

void
handle_pending_retransmit(bcast_cache *cache) 
{ cache->send();
}

/************************************************************************
 * Methods dealing with the complete packet CACHE
 ************************************************************************/

// add an entry to the cache, provided it there is space
// (otherwise, delete oldest entry first)

void bcast_cache::add_to_cache(char *msg, int msg_len, in_addr_t addr, u_int16_t port,
                               struct bcastHeader msg_hdr)
{ 
#ifdef DEBUG
  fprintf(stderr, "BCAST: add packet (%3d.%3d.%3d.%3d:%d) to packet cache\n",
    NIPQUAD(msg_hdr.sender_addr), msg_hdr.sequence_no);
#endif

  // check whether buffer is full, if so, delete first (oldest) entry
  if (cache.size() == (u_int32_t)bcast_obj->buffer_size) {
#ifdef DEBUG
    fprintf(stderr, "BCAST: delete first entry in packet cache\n");
#endif
    cache.erase(cache.begin());
  }

  // we are done: create an entry and add to the end of the CACHE
  bcast_cache_entry *new_entry = new bcast_cache_entry(msg,msg_len,addr,port,msg_hdr);
  cache.insert(cache.end(), *new_entry);
}

void bcast_cache::init_send(in_addr_t src, u_int32_t seq_no)
{ list<bcast_cache_entry>::iterator cache_iter;
  double delay = PENDING_DELAY * (double)(random())/RAND_MAX;
  for (cache_iter = cache.begin(); cache_iter != cache.end(); cache_iter++) {
    if ((*cache_iter).bcast_hdr.sender_addr == src &&
        (*cache_iter).bcast_hdr.sequence_no == seq_no) {
#ifdef DEBUG
      fprintf(stderr, "BCAST: schedule (%3d.%3d.%3d.%3d:%d) for retransmit\n", 
        NIPQUAD(src), seq_no);
#endif
      // save current packet, so even if cache changes, once timer
      // goes off, we still have a copy of the packet
      pending = new bcast_cache_entry((*cache_iter).payload, 
          (*cache_iter).payload_len, (*cache_iter).dst_addr,
          (*cache_iter).dst_port, (*cache_iter).bcast_hdr); 

      // set retransmission timer (delays retransmissions, chance to
      // overhear other retransmissions and cancel ours
      bcast_obj->retransmit_t = bcast_obj->bcast_events->new_oneoff_after_ms
           ((int)delay, callback(handle_pending_retransmit, this));
      return;
    }
  }
}

void bcast_cache::send()
{ unsigned char msg[BCAST_MAX_MSG];

  // we have to change the entry on BCAST_HDR to list the current node
  // as previous hop
  pending->bcast_hdr.prev_hop = bcast_obj->sender_addr;

#ifdef DEBUG
  fprintf(stderr, "BCAST: send delayed NACK'd packet (%3d.%3d.%3d.%3d, %d)\n",
    NIPQUAD(pending->bcast_hdr.sender_addr), pending->bcast_hdr.sequence_no);
#endif

  // assemble message: first BCAST_HDR, then PAYLOAD
  bcopy((char *)&pending->bcast_hdr, msg, sizeof(bcastHeader));
  bcopy(pending->payload, msg + sizeof(struct bcastHeader), 
    pending->payload_len);

  // send to correct destination addr
  bcopy(&(pending->dst_addr), &(bcast_obj->destAddr.sin_addr), 
    sizeof(pending->dst_addr));
  bcast_obj->destAddr.sin_port = pending->dst_port;

  int rc = sendto(bcast_obj->bcast_out_sock, msg,
                  pending->payload_len + sizeof(struct bcastHeader), 0,
                  (struct sockaddr *) &(bcast_obj->destAddr),
                  sizeof(bcast_obj->destAddr));
  if (rc < 0) {
    fprintf(stderr, "BCAST: cannot send delayed NACK'd multicast packet\n");
    exit(1);
  }

  // release temporary storage for pending message
  free(pending);
  pending = NULL;
}

void bcast_cache::cancel_send(in_addr_t src, u_int32_t seq_no)
{ // check whether there is a pending timer
  if (bcast_obj->retransmit_t.scheduled()) {
    // check whether the pending packet matches the one currently heard
    if (pending->bcast_hdr.sender_addr == src &&
        pending->bcast_hdr.sequence_no == seq_no) {
#ifdef DEBUG
  fprintf(stderr, "BCAST: cancel pending transmission of NACK'd packet");
  fprintf(stderr, " (%3d.%3d.%3d.%3d, %d)\n",
    NIPQUAD(pending->bcast_hdr.sender_addr), pending->bcast_hdr.sequence_no);
#endif
      // unschedule timer, release temporary storage for pending packet
      bcast_obj->retransmit_t.unschedule();
      free(pending);
      pending = NULL;
    }
  }
}
