/*******************************************************************
 Copyright (C) 2004 Thomas Kunz, CRC Canada, BCAST for IPv4.
 DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 See the GNU Library General Public License (file COPYING in
 directory application) for conditions of use and redistribution.
*********************************************************************/

/*
 * Functions to manage neighbor info. We learn about neighbors from
 * periodic HELLO messages, neighbors (and their neighbors) are saved
 * in a list. If the information is not renewed, it will expire soon.
 */

#include "bcast_module.h"
#include "bcast_tgt.hh"

#define CURRENT_TIME(tv)    (void) gettimeofday(&tv, NULL)
#define DEBUG

void
handle_neighbors(XrlBcastTarget *obj, Neighbors *nbs) 
{ // schedule a new purge of expired neighbor info after missing
  // ALLOWED_HELLO_LOSS HELLO intervals
  int delay = obj->hello_interval * ALLOWED_HELLO_LOSS * 1000; // in ms
  obj->neighbor_t = obj->bcast_events->new_oneoff_after_ms(delay,
                   callback(handle_neighbors, obj, nbs));
  nbs->nb_purge();
}

/*
   Neighbor Management  Functions
*/

/* Insert and entry for neighbor with address ADDR. If such a neighbor
 * already exist, update expiration time to CURRENT_TIME + EXPIRY (in seconds)
 */

void
Neighbors::nb_insert(in_addr_t addr, list<in_addr_t> nbs, int expiry) 
{ // Search the list for the neighbor
  list<Neighbor_entry>::iterator iter;
  for (iter  = neighbors.begin(); iter != neighbors.end(); iter++) {
    if ((*iter).nb_addr == addr)
      break;
  }
  if (iter != neighbors.end())
  { // update expiration time and set of neighbors for that entry
    Neighbor_entry *b = &(*iter);
    b->two_hop_nbs = nbs;
    CURRENT_TIME(b->expire);
    b->expire.tv_sec += expiry;
  } else {
    Neighbor_entry *b = new Neighbor_entry(addr, nbs);
    CURRENT_TIME(b->expire);
    b->expire.tv_sec += expiry;
    neighbors.insert(neighbors.end(), *b);
  }
}

/* look for a neighbor with address ADDR and return its list of neighbors
 * (which are two-hop neighbors from the perspective of the current node).
 * If no such neighbour exists, return an empty list.
 */

list<in_addr_t>
Neighbors::nb_lookup(in_addr_t addr) 
{ // Search the list for a match of source and bid
  list<Neighbor_entry>::iterator iter;
  for (iter  = neighbors.begin(); iter != neighbors.end(); iter++) {
    if ((*iter).nb_addr == addr) {
      return (*iter).two_hop_nbs;
    }
  }
  // create an empty list and return
  list<in_addr_t> empty;
  empty.clear();
  return empty;
}

/* Return all one-hop neighbors of the current node
 */

list<in_addr_t> 
Neighbors::one_hop_nbs()
{ list<in_addr_t> nbs;
  list<Neighbor_entry>::iterator iter;
  for (iter  = neighbors.begin(); iter != neighbors.end(); iter++) {
    nbs.insert(nbs.end(), (*iter).nb_addr);
  }
  return nbs;
}

/* return the maximum node degree for any neighbor node, including this node
 */

uint
Neighbors::max_degree()
{ uint degree = node_degree(); // number of neighbors of node itself
  list<Neighbor_entry>::iterator iter;
  for (iter  = neighbors.begin(); iter != neighbors.end(); iter++) {
    if ((*iter).two_hop_nbs.size() > degree)
      degree = (*iter).two_hop_nbs.size();
  }
  return degree;
}

/* return the number of neighbors for this node
 */

uint
Neighbors::node_degree()
{ return neighbors.size();
}

/* periodically purge neighbor information that has gone stale (i.e., has not
 * been refreshed by a recent HELLO message), as we assume that the neighbor
 * does not exist anymore. The periodicity is determined by the constant 
 * ALLOWED_HELLO_LOSS, for details see the helper function HANDLE_NEIGHBORS
 * above.
 */

void
Neighbors::nb_purge() {
  struct timeval current_time;
  CURRENT_TIME(current_time);

#ifdef DEBUG
  fprintf(stderr, "BCAST: purging neighbor information\n");
#endif

  list<Neighbor_entry>::iterator iter, tmp_iter;
  iter = neighbors.begin();
  while (iter != neighbors.end()) {
    tmp_iter = iter;
    tmp_iter++; // save pointer to next element
    if (((*iter).expire.tv_sec < current_time.tv_sec) 
	|| ((*iter).expire.tv_sec == current_time.tv_sec && 
            (*iter).expire.tv_usec < current_time.tv_usec)) {
      neighbors.erase(iter); // this invalidates the iterator
    }
    iter = tmp_iter;
  }
}

